import timeit
import numpy as np
from time import time as now
import sys


def time_once(code, n=1):
    ''' Measures execution time of code (best of 3).

    :param code: Code string or callable.
    :param n: Number of times to repeat execution (n x 3).

    :returns: The best execution time.
    '''
    times = min(timeit.Timer(code).repeat(3, n))

    return np.array(times)/n


class LRUCache():
    ''' Given a function and LRU caching implementation,
        caches the results of the function.

        :ivar impl: LRU cache implementation, for instance \
            `functools.lru_cache`.
        :ivar func: The function to cache.
        :ivar maxsize: The size of the cache.
        :ivar typed: Determines whether a distinction is \
            made between arguments of different types.
        :ivar cached: The cached function.

        >>> from dautil import perf
        >>> from functools import lru_cache
        >>> cache = perf.LRUCache(lru_cache, lambda x: x)
        >>> cache.cache()
        >>> cache.cached(1)
        1
        >>> cache.cached(1)
        1
        >>> cache.hits_miss()
        1.0
    '''
    def __init__(self, impl, func,
                 maxsize=128, typed=False):
        '''

        :param impl: LRU cache implementation, for instance \
            functools.lru_cache.
        :param func: The function to cache.
        :param maxsize: The size of the cache.
        :param typed: Determines whether a distinction is \
            made between arguments of different types.
        '''
        self.impl = impl
        self.func = func
        self.maxsize = maxsize
        self.typed = typed
        self.cached = None
        self.info = None

    def cache(self):
        ''' Caches the function.  '''
        self.cached = self.impl(self.maxsize,
                                self.typed)(self.func)

    def clear(self):
        ''' Clears the cache. '''
        if self.cached:
            self.cached.cache_clear()
            self.info = None

    def get_info(self):
        ''' Gets cache info. '''
        if self.cached:
            self.info = self.cached.cache_info()

    def hits_miss(self):
        ''' Calculates hits/miss ratio.
            In a muti-threaded environment, the calculation is approximate.

        :returns: The hits/miss ratio.
        '''
        if self.info is None:
            self.get_info()

        if self.info.misses == 0:
            return None

        return self.info.hits/self.info.misses


class StopWatch():
    ''' A simple stopwatch, which has a context manager.

    :ivar elapsed: Elapsed time in seconds.

    >>> from dautil import perf
    >>> with perf.StopWatch() as sw:
    ...     pass
    ...
    >>> sw.elapsed
    7.867813110351562e-06
    '''
    def __enter__(self):
        self.begin = now()
        self.elapsed = 0
        return self

    def __exit__(self, *args):
        self.elapsed = now() - self.begin


class CountMinSketch():
    ''' CountMin sketch implementation.

    :ivar depth: Depth of the CountMin sketch table.
    :ivar width: Width of the CountMin sketch table.
    :ivar table: The CountMin sketch table.
    '''
    def __init__(self, depth=5, width=20, seed=28):
        self.depth = depth
        self.width = width
        np.random.seed(seed)
        self.table = np.zeros((depth, width))
        self.hashes = np.random.random_integers(0, sys.maxsize - 1, depth)
        self.PRIME_MODULUS = (1 << 31) - 1

    def hash(self, item, i):
        ''' Calculates the hash for a given item.

        :param item: An item.
        :param i: The index for which to compute the hash.

        :returns: The hash for the item.
        '''
        hval = self.hashes[i] * item
        hval += hval >> 32
        hval &= self.PRIME_MODULUS

        return int(hval) % self.width

    def add(self, item):
        ''' Adds an item.

        :param item: An item.
        '''
        for i in range(self.depth):
            self.table[i][self.hash(item, i)] += 1

    def estimate_count(self, item):
        ''' Estimates the count for an item.

        :param item: An item.

        :returns: The estimated count.
        '''
        res = sys.maxsize

        for i in range(self.depth):
            res = min(res, self.table[i][self.hash(item, i)])

        return res



""" Logging utilities. """
from pkg_resources import get_distribution
from pkg_resources import resource_filename
from pkg_resources import DistributionNotFound
import logging
import logging.config
import pprint
from appdirs import AppDirs
import os
from dautil import conf
from dautil import options
import inspect


def env_logger():
    """ Creates a configurable logger.
    The environment variable **DAUTIL_LOGGER** controls
    whether the logger is enabled.
    If the value of the variable is *, then all loggers
    will be enabled, otherwise the calling function name
    is matched against `DAUTIL_LOGGER`.
    Enabling/disabling the logger may be influenced by
    an underlying framework.

    :returns: A configurable logger.
    """
    frame = inspect.stack()[1]
    func_name = frame[3]
    env_val = os.environ.get('DAUTIL_LOGGER')
    log_enabled = False

    if env_val:
        if env_val == '*':
            log_enabled = True
        elif func_name in env_val:
            log_enabled = True

    logger = conf_logger(func_name)

    if not log_enabled:
        logger.addHandler(logging.NullHandler())
        # Needed for ipython notebook sessions
        logger.propagate = False

    return logger


def get_logger(name):
    """ Creates a logger using an internal configuration file.

    :param name: The name of the logger.

    :returns: The configured logger.
    """
    log_config = resource_filename(__name__, 'log.conf')
    logging.config.fileConfig(log_config)
    logger = logging.getLogger(name)

    return logger


def conf_logger(name):
    """ Creates a logger using a configuration file
    provided by the user in the current working directory.
    If a configuration file is not found, dautil uses
    basic configuration.

    :param name: The name of the logger.

    :returns: The configured logger.
    """
    # TODO check for conf file in appropriate dir
    conf_file = 'dautil_log.conf'

    if conf.file_exists(conf_file):
        logging.config.fileConfig(conf_file, disable_existing_loggers=False)
    else:
        logging.basicConfig()

    return logging.getLogger(name)


def shorten(module_name):
    """ Helper function which shortens a module name
    using the first occurrence of a dot. For example
    pandas.io.wb is shortened to pandas.

    :param module_name: The name of the module.

    :returns: The shortened name.

    >>> from dautil import log_api
    >>> log_api.shorten('pandas.io.wb')
    'pandas'
    """
    dot_i = module_name.find('.')

    return module_name[:dot_i]


def log(modules, name):
    """ Logs the versions of imported modules in a
    best effort fashion. Some common modules are excluded.

    :param modules: A list of modules as available in sys.modules.
    :param name: The name of the logger.

    :returns: A dictionary with modules as keys and versions as values.
    """
    skiplist = ['pkg_resources', 'distutils', 'zmq']

    logger = get_logger(name)
    logger.debug('Inside the log function')
    versions = {}

    for k in modules.keys():
        str_k = str(k)

        if '.version' in str_k:
            short = shorten(str_k)

            if short in skiplist:
                continue

            try:
                ver = get_distribution(short).version
                logger.info('%s=%s' % (short, ver))
                versions[short] = ver
            except ImportError:
                logger.warn('Could not import', short)
            except DistributionNotFound as e:
                logger.warn('Could not find distribution %s', e)

    return versions


class VersionsLogFileHandler(logging.FileHandler):
    """ A specialized file handler for the logging of software versions. """
    def __init__(self, fName):
        dirs = AppDirs(options.APP_DIR, "Ivan Idris")
        path = dirs.user_log_dir
        print(path)

        if not os.path.exists(path):
            os.mkdir(path)

        super(VersionsLogFileHandler, self).__init__(
              os.path.join(path, fName))


class Printer():
    """ Utility class for pretty printing.

    :ivar nelems: Max number of elements in a list to print.
    """
    def __init__(self, modules=None, name=None, nelems=-1):
        if modules and name:
            log(modules, name)

        self.nelems = nelems

    def print(self, *args, **kwargs):
        """ Pretty prints a message. """
        for arg in args:
            if self.nelems > 0 and isinstance(arg, list):
                pprint.pprint(self.compress_mid(arg), **kwargs)
            else:
                pprint.pprint(arg, **kwargs)

        print()

    def compress_mid(self, alist, replace='...'):
        """ Takes the head, mid and the tail of a list \
            and discards the rest. For example

        >>> from dautil import log_api
        >>> p = log_api.Printer(nelems=3)
        >>> p.compress_mid(list(range(5)))
        [0, '...', 2, '...', 4]

        :param alist: A list.
        :param replace: A string used to replace discarded elements.

        :returns: The list with discarded values replaced.
        """
        compressed = alist.copy()
        N = len(alist)

        if N > self.nelems:
            half = self.nelems//2
            mid = N//2
            logger = env_logger()
            compressed = alist[:half]

            for i in range(half, N - half):
                if i != mid and compressed[-1] != replace:
                    compressed.append(replace)
                    logger.debug('Compressed %s', compressed)
                if i == mid:
                    compressed.append(alist[mid])

            compressed.extend(alist[-half:])

        return compressed



""" IPython/Jupyter notebook widgets and utilities. """
from IPython.display import display
from IPython.display import Math
from IPython.html import widgets
from dautil import collect
from dautil import conf
from dautil import ts
from dautil import log_api
import matplotlib as mpl
from matplotlib.colors import rgb2hex
import pprint
from matplotlib.colors import ColorConverter


def create_month_widget(month, *args, **kwargs):
    """ Creates a dropdown wiget with short month names
    as labels.

    :param month: The month to select by default.

    :returns: The configured month widget.
    """
    return widgets.Dropdown(options=ts.short_months(),
                            selected_label=month, *args, **kwargs)


class WidgetFactory():
    ''' A factory for IPython widgets part of the \
        `RcWidget` GUI.

        :ivar rc_widget: A `RcWidget` instance.
    '''
    def __init__(self, rcw):
        self.rc_widget = rcw

    # TODO allow ‘rgbcmykw’
    # TODO allow standard names like 'aqua'
    def color_chooser(self, property):
        """ Creates a box with widgets related to choosing a color.

        :param property: A color related key in matplotlib.rcParams.

        :returns: A box with widgets.
        """
        cc = ColorConverter()
        rgb = cc.to_rgb(mpl.rcParams[property])
        logger = log_api.env_logger()
        logger.debug('{0} {1}'.format(property, rgb))

        r = widgets.FloatSlider(min=0, max=1, value=rgb[0], description='Red')
        r.border_color = 'red'

        g = widgets.FloatSlider(min=0, max=1, value=rgb[1],
                                description='Green')
        g.border_color = 'green'

        b = widgets.FloatSlider(min=0, max=1, value=rgb[2], description='Blue')
        b.border_color = 'blue'

        h = widgets.widget_string.HTML(property)
        # TODO put this in a func
        hex = rgb2hex((rgb[0], rgb[1], rgb[2]))
        h.value = '<p style="background-color: {0};">{0}</p>'.format(hex)

        def update(name, value):
            hex = rgb2hex((r.value, g.value, b.value))
            h.value = '<p style="background-color: {0};">{0}</p>'.format(hex)
            self.rc_widget.process(property, hex)

        r.on_trait_change(update, 'value')
        g.on_trait_change(update, 'value')
        b.on_trait_change(update, 'value')

        box = widgets.VBox(children=(r, g, b, h))
        box.border_style = 'dotted'
        box.description = property

        return box


class PageBuilder():
    """ Creates a page with widgets for the `RcWidget`.

        :ivar widgets: A dictionary containing widgets.
        :ivar prefix: The prefix for the widget, for instance 'axes.'.
        :ivar factory: A `WidgetFactory`.
        'ivar keys: A list of matplotlib properties.
    """
    def __init__(self, prefix, factory):
        self.widgets = {}
        self.prefix = prefix
        self.factory = factory
        self.keys = collect.filter_dict_keys(
            lambda x: x.startswith(self.prefix), mpl.rcParams)
        self.add_color_choosers()

    def add(self, widget):
        """ Adds a new widget to the internal dictionary.

        :param widget: An IPython HTML widget.
        """
        self.widgets[widget.description] = widget

    def add_color_choosers(self):
        """ Adds color choosers for relevant properties starting
        with a prefix such as axes.
        """
        logger = log_api.env_logger()
        logger.debug('self.keys {}'.format(self.keys))
        color_keys = collect.filter_list(lambda x: x.endswith('color'),
                                         self.keys)
        logger.debug('Color keys {}'.format(color_keys))

        for key in color_keys:
            self.widgets[key] = self.factory.color_chooser(key)

    def build(self):
        """ Builds an `Accordion` containing widgets.

        :returns: The `Accordion` with widgets sorted by descriptions.
        """
        self.widgets = collect.sort_dict_by_keys(self.widgets)
        box = widgets.Accordion()
        box.children = [self.widgets[k] for k in self.widgets.keys()]

        for i, k in enumerate(self.widgets.keys()):
            box.set_title(i, k)

        return box


# TODO mix DIY widgets with WidgetFactory calls
class RcWidget():
    """ This widget configures the
    `matplotlib.rcParams` global settings.

    :ivar context: A `Context` instance.
    :ivar factory: A `WidgetFactory` instance.
    """
    def __init__(self, context=None):
        self.context = context
        self.factory = WidgetFactory(self)

        if self.context:
            rc = self.context.read_rc()

            if rc:
                mpl.rcParams.update(rc)

        self.old = mpl.rcParams.copy()

        tab = widgets.Tab(children=[self.axes_page(), self.figure_page(),
                                    self.font_page(), self.grid_page(),
                                    self.lines_page()])
        tab.set_title(0, 'Axes')
        tab.set_title(1, 'Figure')
        tab.set_title(2, 'Font')
        tab.set_title(3, 'Grid')
        tab.set_title(4, 'Lines')
        display(tab)

        self.updates_text = widgets.HTML()
        display(self.updates_text)

        self.params_text = widgets.HTML()
        display(self.params_text)

        self.show_params = widgets.widget_button.Button()
        self.show_params.description = 'Show rcParams'
        self.show_params.on_click(self.print_params)
        display(self.show_params)

    def print_params(self, button_instance):
        """ Prints the current matplotlib.rcParams in a textarea.

        :param button_instance: The button to click on.
        """
        html = '<textarea rows="5" cols="50" readonly>{}</textarea>'
        self.params_text.value = html.format(pprint.pformat(mpl.rcParams))

    def process(self, param, value):
        """ Processes changes to the GUI and updates `matplotlib.rcParams`.

        :param param: A key in the `matplotlib.rcParams` dictionary.
        :param value: A value in the `matplotlib.rcParams` dictionary.
        """
        logger = log_api.env_logger()
        logger.debug('name={0}, value={1}'.format(param, value))
        self.params_text.value = ''
        mpl.rcParams[param] = value
        updates = collect.dict_updates(self.old, mpl.rcParams)

        if self.context:
            self.context.update_rc(updates)

        self.updates_text.value = ('<p>mpl.RcParams updates {}</p>'.
                                   format(updates))

    def axes_page(self):
        """ Creates a tab page for the `matplotlib.rcParams`
        keys which start with **axes.**"""
        linewidth = create_linewidth_slider('axes.linewidth')

        titlesize = create_size_slider('axes.titlesize')

        def update_axes_linewidth(name, value):
            self.process(linewidth.description, value)

        def update_titlesize(name, value):
            self.process(titlesize.description, value)

        linewidth.on_trait_change(update_axes_linewidth, 'value')
        titlesize.on_trait_change(update_titlesize, 'value')

        page = PageBuilder('axes.', self.factory)
        page.add(linewidth)
        page.add(titlesize)

        return page.build()

    def font_page(self):
        """ Creates a tab page for the `matplotlib.rcParams`
        keys which start with **font.**"""
        size = create_size_slider('font.size')

        def update_font_size(name, value):
            self.process(size.description, value)

        size.on_trait_change(update_font_size, 'value')

        page = PageBuilder('font.', self.factory)
        page.add(size)

        return page.build()

    def figure_page(self):
        """ Creates a tab page for the `matplotlib.rcParams`
        keys which start with **figure.**"""
        figsize = widgets.Box()
        figsize.description = 'figure.figsize'
        figsize_val = mpl.rcParams[figsize.description]
        height = widgets.FloatSlider(min=0, max=16,
                                     value=figsize_val[0],
                                     description='Height')
        width = widgets.FloatSlider(
            min=0, max=12, value=figsize_val[1],
            description='Width')
        figsize.children = [height, width]

        def update_fig_size(name, value):
            self.process(figsize.description, (height.value, width.value))

        height.on_trait_change(update_fig_size, 'value')
        width.on_trait_change(update_fig_size, 'value')
        page = PageBuilder('figure.', self.factory)
        page.add(figsize)

        return page.build()

    def grid_page(self):
        """ Creates a tab page for the `matplotlib.rcParams`
        keys which start with **grid.**"""
        logger = log_api.env_logger()
        logger.debug('Created grid page')
        linewidth = create_linewidth_slider('grid.linewidth')

        def update_linewidth(name, value):
            self.process(linewidth.description, value)

        linewidth.on_trait_change(update_linewidth, 'value')

        page = PageBuilder('grid.', self.factory)
        page.add(linewidth)

        return page.build()

    def lines_page(self):
        """ Creates a tab page for the `matplotlib.rcParams`
        keys which start with **lines.**"""
        linewidth = create_linewidth_slider('lines.linewidth')

        def update_linewidth(name, value):
            self.process(linewidth.description, value)

        linewidth.on_trait_change(update_linewidth, 'value')

        page = PageBuilder('lines.', self.factory)
        page.add(linewidth)

        return page.build()


def create_linewidth_slider(desc, *args, **kwargs):
    """ Creates a slider for linewidth-type settings
    in `matplotlib.rcParams`.

    :param desc: The description label of the widget.

    :returns: The configured slider.
    """
    from_rc = mpl.rcParams[desc]

    val = 0

    # TODO deal with strings
    if not isinstance(from_rc, str):
        val = from_rc

    return widgets.IntSlider(min=0, max=9, value=val,
                             description=desc, *args, **kwargs)


def create_size_slider(desc, *args, **kwargs):
    """ Creates a slider for size-type settings
    in `matplotlib.rcParams`.

    :param desc: The description label of the widget.

    :returns: The configured slider.
    """
    from_rc = mpl.rcParams[desc]

    val = 0

    # TODO deal with strings
    if not isinstance(from_rc, str):
        val = from_rc

    return widgets.FloatSlider(min=0, value=val,
                               description=desc, *args, **kwargs)


class LatexRenderer():
    """ Utility class which helps number and render Latex
    in a IPython/Jupyter notebook.

    :ivar chapter: Chapter number.
    :ivar curr: Current equation number.
    :ivar numbers: List of used equation numbers.
    :ivar context: A `Context` instance.

    .. code-block:: python

        import dautil as dl
        lr = dl.nb.LatexRenderer(chapter=6, start=6, context=context)
        lr.render(r'Y_j= \sum _{i=-(m-1)/2}')
    """
    def __init__(self, chapter=None, start=1, context=None):
        self.chapter = chapter
        self.curr = start
        self.numbers = []
        self.context = context

        if self.context:
            from_context = self.context.read_latex()

            if from_context:
                log_api.Printer().print(from_context)
                eqn_list = list(from_context.values())
                assert start not in collect.flatten(eqn_list), from_context

    # DIY numbering because IPython doesn't
    # support numbering
    def number_equation(self):
        """ Creates a Latex string relating
        to the numbering of equations.

        :returns: A Latex string with the correct equation number.
        """
        number = '('

        if self.chapter:
            number += str(self.chapter) + '.'

        number += str(self.curr) + ')\hspace{1cm}'

        return number

    def render(self, equation):
        """ Renders an equation.

        :param equation: A string containing the equation.
        """
        number = self.number_equation()
        self.numbers.append(self.curr)
        logger = log_api.env_logger()

        if self.context:
            logger.debug(self.numbers)
            self.context.update_latex(self.numbers)

        display(Math(r'%s' % (number + equation)))
        self.curr += 1


# TODO store key/id information in sql lite db in CONF_DIR
# with key mnemonic, creation time etc
class Context():
    """ A mediator for the storing and retrieving
    of configuration settings.

    :ivar fname: Name of the context, this should be unique \
        such as the name of a notebook
    """
    def __init__(self, fname):
        self.fname = fname
        self.labels = fname + '.labels'
        self.latex = fname + '.latex'

    def read_rc(self):
        """ Reads the current configuration
        settings related to `matplotlib.rcParams`,
        which are used by `RcWidget`.

        :returns: The current configuration settings or \
        an empty dict.
        """
        config = conf.read_rc()

        if config:
            config = config.get(self.fname, {})

        logger = log_api.env_logger()
        logger.debug('config %s', config)

        return config

    def update_rc(self, updates):
        """ Updates the configuration settings related to
        `matplotlib.rcParams` used by `RcWidget`.

        :param updates: Changes to the configuration.
        """
        conf.update_rc(self.fname, updates)

    def read_labels(self):
        """ Reads the current configuration settings related
        to the `matplotlib.rcParams` used by `LabelWidget`.

        :returns: The current configuration settings\
        or None if no settings are found.
        """
        config = conf.read_rc()

        if config:
            config = config.get(self.labels, None)

        return config

    def update_labels(self, updates):
        """ Updates the configuration settings related to
        `matplotlib.rcParams` used by `LabelWidget`.

        :param updates: Changes to the configuration.
        """
        conf.update_rc(self.labels, updates)

    def read_latex(self):
        """ Reads the current configuration settings related
        to the `LatexRenderer`.

        :returns: The current configuration settings\
        or None if no settings are found.
        """
        config = conf.read_rc()

        if config:
            keys = collect.filter_dict_keys(lambda x: x.endswith('.latex'),
                                            config)
            config = collect.dict_from_keys(config, keys)
            config.pop(self.latex, None)

        return config

    def update_latex(self, updates):
        """ Updates the configuration settings related to
        `LatexRenderer`.

        :param updates: Changes to the configuration.
        """
        conf.update_rc(self.latex, updates)


class NullContext(Context):
    """ A context following the Null Object Pattern
    which does nothing """
    def __init__(self):
        pass

    def __bool__(self):
        return False

    def read_rc(self):
        pass

    def update_rc(self, updates):
        pass

    def read_labels(self):
        pass

    def update_labels(self, updates):
        pass

    def read_latex(self):
        pass

    def update_latex(self, updates):
        pass


class LabelWidget():
    """ A widget you can use to easily fill
    in strings for titles, xlabels and ylabels
    of matplotlib subplots.

    :ivar context: A `Context` instance.
    :ivar labels: A grid of labels.

    .. code-block:: python

        import dautil as dl
        dl.nb.LabelWidget(2, 2, context)
    """
    def __init__(self, nrows=1, ncols=1, context=NullContext()):
        assert context, 'Define context'
        self.context = context
        self.labels = collect.GridList(nrows, ncols, {})
        self.read_old_labels()

        for i in range(nrows):
            children = []

            for j in range(ncols):
                labels_box = self.create_mpl_labels_box(i, j,
                                                        self.labels.grid[i][j])
                children.append(labels_box)

            display(widgets.HBox(children=children))
            display(widgets.HTML('<br/>'))

    def read_old_labels(self):
        """ Reads the labels from a configuration file. """
        old = self.context.read_labels()

        if old:
            self.labels.fill(old)

    def update(self, name, value, row, col):
        """ Updates an internal data structure and
        related configuration file.

        :param name: title, xlabel, legend or ylabel.
        :param value: A string representing a label. \
        If needed use curly braces as used by the Python \
        format() string method.
        :param row: The number of the row.
        :param col: The number of the col.
        """
        self.labels.update(row, col, {name: value})
        self.context.update_labels(self.labels.grid)

    def create_mpl_labels_box(self, row, col, old):
        """ Creates a box with the widgets for a single
        subplot (cell).

        :param row: The row number of the subplot.
        :param col: The column number of the subplot.
        :param old: The setting for this subplot from a configuration file.

        :returns: The box with widgets.
        """
        box = widgets.VBox()
        box.border_color = 'red'
        coord = ' [{0}][{1}]'.format(row, col)
        title = widgets.widget_string.Text(old.get('title', ''),
                                           description='title' + coord)
        xlabel = widgets.widget_string.Text(old.get('xlabel', ''),
                                            description='xlabel' + coord)
        ylabel = widgets.widget_string.Text(old.get('ylabel', ''),
                                            description='ylabel' + coord)
        legend = widgets.Dropdown(options=['No Legend', 'loc=best',
                                           'loc=upper right',
                                           'loc=upper left'],
                                  selected_label=old.get('legend',
                                                         'No Legend'))
        box.children = [title, xlabel, ylabel, legend]

        def update_title(name, value):
            self.update('title', value, row, col)

        title.on_trait_change(update_title, 'value')

        def update_xlabel(name, value):
            self.update('xlabel', value, row, col)

        xlabel.on_trait_change(update_xlabel, 'value')

        def update_ylabel(name, value):
            self.update('ylabel', value, row, col)

        ylabel.on_trait_change(update_ylabel, 'value')

        def update_legend(name, value):
            self.update('legend', value, row, col)

        legend.on_trait_change(update_legend, 'value')

        return box



""" Utilities to download and load data. """
import pandas as pd
from pandas.io import wb
from pandas.io.data import DataReader
import numpy as np
from io import BytesIO
import zipfile
import gzip
from appdirs import AppDirs
import os
from pkg_resources import resource_filename
import urllib.request as urlrequest
from dautil import log_api
from dautil import options
from collections import namedtuple
from joblib import Memory
import csv
from decimal import Decimal
from datetime import datetime
import pickle


def from_pickle(fname):
    ''' Loads object from pickle.

    :param fname: The name of the pickle file.

    :returns: The object from the pickle.
    '''
    pkl = open(fname, 'rb')
    obj = pickle.load(pkl)
    pkl.close()

    return obj


def centify(text, multiplier=100):
    ''' Converts a string representing money\
        to the corresponding number in cents.

    :param text: A string such as 10.55.
    :param multiplier: A multiplier for the conversion.
    :returns: Cents as an integer, for instance 1055.

    >>> from dautil import data
    >>> data.centify('10.55')
    1055
    '''
    return int(multiplier * Decimal(text))


def dropinf(arr):
    """ Removes `np.inf` and `np.nan` values.

    :param arr: Array with numbers.

    :returns: The cleaned array.

    >>> from dautil import data
    >>> import numpy as np
    >>> arr = np.array([np.inf, 0, 42, np.nan])
    >>> data.dropinf(arr)
    array([  0.,  42.])
    """
    return arr[np.isfinite(arr)]


def download(url, out):
    """ Download a file from the web.

    :param url: The URL of the file.
    :param out: The path of the file.
    """
    req = urlrequest.Request(url, headers={'User-Agent': 'Mozilla/5.0'})
    res = urlrequest.urlopen(req)

    with open(out, 'wb') as data_file:
        data_file.write(res.read())
        log_api.conf_logger(__name__).warning('Downloaded ' + data_file.name)


def process_gzip(url, file_path):
    """ Downloads and uncompresses a GZIP file.

    :param url: The URL of the GZIP file.
    :param file_path: The path of the file.
    """
    response = urlrequest.urlopen(url)
    compressed = BytesIO(response.read())
    decompressed = gzip.GzipFile(fileobj=compressed)

    with open(file_path, 'wb') as outfile:
        outfile.write(decompressed.read())


def process_zip(url, path, fname):
    """ Downloads and uncompresses a ZIP file.

    :param url: The URL of the ZIP file.
    :param path: The path of the file.
    :param fname: The name of the file.

    :returns: The contents of the extracted file.
    """
    response = urlrequest.urlopen(url)

    zipf = zipfile.ZipFile(BytesIO(response.read()))

    return zipf.extract(fname, path)


def get_data_dir():
    """ Finds the appropriate data directory to store data files.

    :returns: A data directory, which is OS dependent.
    """
    dirs = AppDirs(options.APP_DIR, "Ivan Idris")
    path = dirs.user_data_dir
    log_api.conf_logger(__name__).warning('Data dir ' + path)

    if not os.path.exists(path):
        os.mkdir(path)

    return path


def get_direct_marketing_csv():
    ''' Retrieves a CSV file with direct marketing data as described in \
        http://blog.minethatdata.com/2008/03/ \
        minethatdata-e-mail-analytics-and-data.html

    :returns: The path to the downloaded file.
    '''
    base_url = 'https://azuremlsampleexperiments.blob.core.windows.net/'
    url = base_url + 'datasets/direct_marketing.csv'
    out = os.path.join(get_data_dir(), 'direct_marketing.csv')

    if not os.path.exists(out):
        download(url, out)

    return out


def get_smashing_baby():
    ''' Retrieves a WAV file of Austin Powers.

    :returns: The path to the downloaded file.
    '''
    base_url = 'http://www.thesoundarchive.com/'
    url = base_url + 'austinpowers/smashingbaby.wav'
    out = os.path.join(get_data_dir(), 'smashingbaby.wav')

    if not os.path.exists(out):
        download(url, out)

    return out


def read_csv(fname):
    ''' Reads a CSV file and returns a list of dictionaries \
        where each line corresponds to a line in the file.

    :param fname: The name or path of the file.

    :returns: The dictionary.
    '''
    csv_rows = []

    with open(fname) as afile:
        reader = csv.DictReader(afile)

        csv_rows = [row for row in reader]

    return csv_rows


class Nordpil():
    """ Utility class to get data from the Nordpil website.

    :ivar dir: The data destination directory.
    """
    def __init__(self):
        self.dir = get_data_dir()

    def load_urban_tsv(self):
        """ Downloads the urbanareas file.

        :returns: The fully qualified path of the downloaded file.
        """
        fname = os.path.join(self.dir,
                             'urbanareas1_1.tsv')

        if not os.path.exists(fname):
            download('http://nordpil.com/static/downloads/urbanareas1_1.tsv',
                     fname)

        return fname


class SPANFB():
    """ Utility class which downloads data from the
    SPAN Facebook webpage.
    :ivar fname: The path of the downloaded file.
    """
    def __init__(self):
        self.fname = os.path.join(get_data_dir(),
                                  'facebook_combined.txt')

    def load(self):
        """ Downloads the SPAN Facebook dataset.

        :returns: The fully qualified path of the downloaded file.
        """
        if not os.path.exists(self.fname):
            process_gzip(
                'https://snap.stanford.edu/data/facebook_combined.txt.gz',
                self.fname)

        return self.fname


class Weather():
    """ Utility class which downloads or loads weather data
    from the KNMI website. """
    @staticmethod
    def beaufort_scale(df):
        """ Categorizes wind speed using the Beaufort scale.
        :param df: A pandas `DataFrame`.

        :returns: A categorized pandas `DataFrame`.
        """
        return pd.cut(df['WIND_SPEED'], bins=[0, 0.3, 1.5, 3.3, 5.5, 8, 10.8,
                                              13.9, 17.2, 20.7, 24.5, 28.4,
                                              32.6, np.inf],
                      labels=np.arange(13, dtype=np.int))

    @staticmethod
    def categorize_wind_dir(df):
        """ Categorize the wind direction (0 - 360) using
        cardinal direction (North, South etc.)

        :param df: A pandas `DataFrame`.

        :returns: A categorized pandas `DataFrame`.
        """
        card_dirs = ['Calm', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE', 'S',
                     'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW', 'N']
        deg_rngs = [0, 11.25, 33.75, 56.25, 78.75, 101.25, 123.75, 146.25,
                    168.75, 191.25, 213.75, 236.25, 258.75, 281.25, 303.75,
                    326.25, 348.75, 360]

        return pd.cut(df['WIND_DIR'], deg_rngs, labels=card_dirs)

    @staticmethod
    def load():
        """ Loads data from an internal pickle.

        :returns: The pandas `DataFrame` loaded from the pickle.
        """
        pckl = resource_filename(__name__, 'weather.pckl')
        return pd.read_pickle(pckl)

    @staticmethod
    def rain_values():
        """ Loads rain values without NA values as a NumPy array.

        :returns: The rain values as a NumPy array.
        """
        return Weather.load()['RAIN'].dropna().values

    @staticmethod
    def get_header(alias):
        """ Gets slightly longer and descriptive column labels.

        :param alias: A short column name.

        :returns: A longer column name.
        """
        mapping = {'WIND_DIR': 'Wind Dir',
                   'WIND_SPEED': 'W Speed, m/s',
                   'TEMP': 'Temp, °C',
                   'RAIN': 'Rain, mm',
                   'PRESSURE': 'Pres, hPa'}

        return mapping.get(alias, 'UNKNOWN')

    @staticmethod
    def get_headers():
        """ Gets the column labels for the pandas `DataFrame`
        stored in the internal pickle.

        :returns: A list that contains the column names.
        """
        return ['WIND_DIR', 'WIND_SPEED', 'TEMP', 'RAIN', 'PRESSURE']

    @staticmethod
    def fetch_DeBilt_weather():
        """ Downloads, cleans and pickles weather data from the
        KNMI website. """
        home = 'http://www.knmi.nl/'
        url = home + 'klimatologie/daggegevens/datafiles3/260/etmgeg_260.zip'
        path = get_data_dir()
        file = process_zip(url, path, 'etmgeg_260.txt')
        df = pd.read_csv(
            file,
            skiprows=47,
            usecols=[
                'YYYYMMDD',
                'DDVEC',
                '   FG',
                '   TG',
                '   RH',
                '   PG'],
            index_col=0,
            parse_dates=True,
            na_values='     ')

        df.columns = ['WIND_DIR', 'WIND_SPEED', 'TEMP', 'RAIN', 'PRESSURE']
        df[df['RAIN'] == -1] = 0.05 / 2
        df['WIND_SPEED'] = 0.1 * df['WIND_SPEED']
        df['TEMP'] = 0.1 * df['TEMP']
        df['RAIN'] = 0.1 * df['RAIN']
        df[df['PRESSURE'] < 1] = np.nan
        df['PRESSURE'] = 0.1 * df['PRESSURE']
        log_api.conf_logger(__name__).warning(df.index[-1])

        pckl = os.path.join(path, 'weather.pckl')
        df.to_pickle(pckl)
        assert df.equals(pd.read_pickle(pckl))


class OHLC():
    ''' Downloads and caches historical EOD data from the web.
        with the `pandas.io.data.DataReader`.
    '''
    def __init__(self, data_source='yahoo'):
        ''' Creates the object and sets up caching.

        :param data_source: A data souce such as `yahoo` or `google`.
        '''
        self.data_source = data_source
        # caching
        memory = Memory(cachedir='.')
        self.get = memory.cache(self.get)

    def get(self, ticker):
        ''' Retrieves EOD data from cache or the web.

        :param ticker: The stock symbol, such as `AAPL`.

        :returns: The data as a pandas `DataFrame`.
        '''
        start = datetime(1900, 1, 1, 0, 0, 0, 0)
        return DataReader(ticker, data_source=self.data_source, start=start)


class Worldbank():
    """ Caching proxy for the pandas Worldbank API.

    :ivar indicators: A list of indicator tuples in the form: \
        `(alias, name, longname)`
    :ivar alias2name: Mapping of alias to name.
    :ivar name2alias: Mapping of name to alias.
    :ivar name2longname: Mapping of name to longname.
    :ivar aliases: A list of aliases.
    :ivar names: A list of indicator names.
    """
    def __init__(self):
        # TODO make this configurable
        Indicator = namedtuple('Indicator', ['alias', 'name', 'longname'])
        self.indicators = [Indicator(alias='pop_grow', name='sp.pop.grow',
                                     longname='Population Growth'),
                           Indicator(alias='gdp_pcap', name='ny.gdp.pcap.cd',
                                     longname='GDP Per Capita'),
                           Indicator(alias='co2', name='en.atm.co2e.kt',
                                     longname='CO2 emissions (kt)'),
                           Indicator(alias='inf_mort', name='sh.dyn.mort',
                                     longname='Mortality rate, \
                                     under-5 (per 1,000 live births)'),
                           Indicator(alias='primary_education',
                                     name='se.prm.cmpt.zs', longname='Primary\
                                     Education Completion Rate')]
        self.alias2name = {i.alias: i.name for i in self.indicators}
        self.name2alias = {j: i for i, j in self.alias2name.items()}
        self.name2longname = {i.name: i.longname for i in self.indicators}
        self.aliases = self.alias2name.keys()
        self.names = self.alias2name.values()

        # caching
        memory = Memory(cachedir='.')
        self.download = memory.cache(self.download)
        self.get_countries = memory.cache(self.get_countries)

    def get_countries(self, *args, **kwargs):
        """ Caches the `pandas.io.wb.get_countries()` results.

        :returns: The result of the query from cache or the WWW.
        """
        return wb.get_countries(*args, **kwargs)

    def download(self, *args, **kwargs):
        """ Caches the `pandas.io.wb.download()` results.

        :returns: The result of the query from cache or the WWW.
        """
        return wb.download(*args, **kwargs)

    def get_alias(self, name):
        """ Gets an internal alias for the official
        Worldbank indicator.

        :param name: The name of the Worldbank indicator.

        :returns: The internal alias.
        """
        assert name in self.names, self.indicators

        return self.name2alias[name]

    def get_name(self, alias):
        """ Gets the official Worldbank indicator
        for an internal alias.

        :param alias: The internal alias.

        :returns: The name of the Worldbank indicator.
        """
        assert alias in self.aliases, self.indicators

        return self.alias2name[alias]

    def get_longname(self, name):
        """ Gets a longer descriptive name for a
        Worldbank indicator.

        :param name: The name of a Worldbank indicator.

        :returns: The long descriptive name.
        """
        assert name in self.names, self.indicators

        return self.name2longname[name]

    def rename_columns(self, df, use_longnames=False):
        """ Renames the columns of a pandas `DataFrame`.

        :param df: A pandas `DataFrame`.
        :param use_longnames: Whether to use longnames for the renaming.

        :returns: The pandas DataFrame with its columns renamed.
        """
        new_cols = []

        for col in df.columns:
            if use_longnames:
                if col in self.names:
                    new_cols.append(self.name2longname[col])
                else:
                    new_cols.append(col)
            else:
                if col in self.names:
                    new_cols.append(self.name2alias[col])
                else:
                    new_cols.append(col)

        df.columns = new_cols

        return df



''' Natural language processing utilities '''
from sklearn.feature_extraction.text import TfidfVectorizer
import numpy as np
import pandas as pd
from dautil import data
from dautil import log_api
import os
import datetime
from nltk.corpus import brown
from nltk.corpus import reuters
from nltk.corpus import names
from joblib import Memory
from nltk.corpus import words


memory = Memory(cachedir='.')


def has_duplicates(str):
    ''' Checks whether a string has repeating words in it.

    :param str: A string.

    :returns: `True` if the string has repeating words else `False`.
    '''
    tokens = str.split()

    return len(tokens) != len(set(tokens))


def has_digits(str):
    ''' Checks whether a string has digits in it.

    :param str: A string.

    :returns: `True` if the string has digits else `False`.
    '''
    for c in str:
        if c.isdigit():
            return True

    return False


def lower_all(alist):
    ''' Lowercases all words/strings in a list

    :param alist: A list of words/strings.

    :returns: The lowercased words.
    '''
    return [w.lower() for w in alist]


@memory.cache
def common_unigrams():
    ''' Gets the unique words in several corpora in a set.

    :returns: The unique words.
    '''
    sw = set(lower_all(brown.words()))
    sw = sw.union(set(lower_all(names.words())))
    sw = sw.union(set(lower_all(words.words())))
    sw = sw.union(set(lower_all(reuters.words())))

    return sw


def calc_tfidf(corpus, sw='english', ngram_range=(2, 3)):
    ''' Calculates TF-IDF for a list of text strings and sums it up by term.

    :param corpus: A list of text strings.
    :param sw: A list of stop words.

    :returns: A pandas `DataFrame` with columns 'term' and 'tfidf'.
    '''
    if ngram_range:
        vectorizer = TfidfVectorizer(ngram_range=ngram_range, stop_words=sw)
    else:
        vectorizer = TfidfVectorizer(stop_words=sw)

    matrix = vectorizer.fit_transform(corpus)
    sums = np.array(matrix.sum(axis=0)).ravel()

    ranks = [(word, val) for word, val in
             zip(vectorizer.get_feature_names(), sums)]

    df = pd.DataFrame(ranks, columns=["term", "tfidf"])
    df = df.sort(['tfidf'])

    return df


def select_terms(df, method='q3', select_func=None):
    ''' Select terms based on TF-IDF.

    :param df: A pandas `DataFrame` as produced by `calc_tfidf` function.
    :param method: The selection method, \
        default use the third quartile as cutoff.
    :param  select_func: An optional selection function.

    :returns: A set containing the selected terms.
    '''
    cutoff = 0

    if method == 'q3':
        cutoff = np.percentile(df['tfidf'], 75)
    else:
        cutoff = select_func(df['tfidf'])

    cut_df = df[df['tfidf'] > cutoff]

    return set(cut_df['term'])


class WebCorpus():
    ''' A corpus for text downloaded from the Web. '''
    def __init__(self, dir):
        self.dir = os.path.join(data.get_data_dir(), dir)
        self.LOG = log_api.conf_logger(__name__)

        if not os.path.exists(self.dir):
            os.mkdir(self.dir)

        self.csv_fname = os.path.join(self.dir, 'metadata.csv')
        self.url_set = set()
        self.init_url_csv()

    def init_url_csv(self):
        ''' Initialize a CSV containing URLs of downloaded texts. '''

        if os.path.exists(self.csv_fname):
            df = pd.read_csv(self.csv_fname)
            self.url_set = set(df['URL'].values.tolist())
        else:
            with open(self.csv_fname, 'w') as urls_csv:
                urls_csv.write('Added,URL,Title,Author\n')

    def store_text(self, name, txt, url, title, author):
        ''' Stores text in the corpus directory. \
            Also updates a CSV file to avoid downloading the \
            same file again.

        :param name: The name of the file.
        :param txt: The text of the file.
        :param url: The URL of the original document.
        :param title: The title of the original document.
        :param author: The author of the original document.
        '''
        if ',' in url:
            return

        fname = os.path.join(self.dir, name.replace('/', ''))

        try:
            with open(fname, 'w') as txt_file:
                txt_file.write(txt)
        except Exception as e:
            self.LOG.warning(e)

        clean_title = title.replace(',', ' ')
        clean_title = clean_title.replace('/', '')

        with open(self.csv_fname, 'a') as urls_csv:
            timestamp = datetime.datetime.now().isoformat()
            urls_csv.write('{0},{1},{2},{3}\n'.format(
                timestamp, url, clean_title,
                author.replace(',', ' ')))

        self.url_set.add(url)

    def get_texts(self):
        ''' Gets all the texts of the corpus.

        :returns: The texts as a list.
        '''
        texts = []

        for f in os.listdir(self.dir):
            if not f.endswith('csv') and not f.startswith('.'):
                try:
                    with open(os.path.join(self.dir, f), 'r') as txt_file:
                        txt = "".join(txt_file.readlines())

                        if len(txt) > 0:
                            texts.append(txt)
                except Exception as e:
                    self.LOG.warning('{0} {1}'.format(f, e))

        return texts

    def get_text(self, name):
        ''' Gets the text for a file in the corpus.

        :param name: The name of the file.
        :param url: The URL of the original document.

        :returns: The text of the file.
        '''
        txt = ''
        fname = os.path.join(self.dir, name)

        with open(fname, 'r') as txt_file:
            txt = "".join(txt_file.readlines())

        return txt

    def get_titles(self):
        df = pd.read_csv(self.csv_fname)
        return set(self.filter(df['Title'].values.tolist()))

    def get_authors(self):
        df = pd.read_csv(self.csv_fname)
        return set(self.filter(df['Author'].values.tolist()))

    def filter(self, str):
        return [i for i in str if isinstance(i, type('s'))]



""" This module contains reporting utilities. """
import pandas as pd
from landslide.generator import Generator as LandslideGenerator
from dautil import data
from dautil import log_api
import os
import sys


def map_grid(a, b, cols, func, *args, **kwargs):
    ''' Apply a function to the cells of a grid.

    :param a: One of two arrays used to form the grid.
    :param b: One of two arrays used to form the grid.
    :param cols: A list of column names.
    :param func: Function to apply.

    :returns: A pandas `DataFrame` containing the result.
    '''
    dfb = DFBuilder(cols=cols)

    for i in a:
        for j in b:
            dfb.row([i, j, func(i, j, *args, **kwargs)])

    return dfb.build().pivot(*cols)


class DFBuilder():
    """ Builds a pandas `DataFrame`,
    by updating a dict incrementally.

    The actual `DataFrame` is built at the end
    of the build process.

    :ivar columns: The columns of the `DataFrame`.
    :ivar df: The `DataFrame`.
    """
    def __init__(self, cols, *args):
        self.columns = cols
        self.df = {}

        for col in self.columns:
            self.df.update({col: []})

        for arg in args:
            self.row(arg)

    def row(self, row):
        """ Adds a row to the pandas DataFrame

        :param row: A list of values to add.

        :returns: A table-like dict.

        >>> from dautil import report
        >>> builder = report.DFBuilder(['A', 'B'])
        >>> row = list(range(2))
        >>> builder.row(row)
        {'B': [1], 'A': [0]}
        >>> builder.df
        {'B': [1], 'A': [0]}
        """
        assert len(row) == len(self.columns),\
            '{0} != expected len {1}'.format(len(row), len(self.columns))

        for col, val in zip(self.columns, row):
            self.df[col].append(val)

        return self.df

    def build(self, index=None):
        """ Builds a pandas `DataFrame` from an internal dict.

        :param index: A list representing the `DataFrame` index.

        :returns: A pandas `DataFrame`.
        """
        self.df = pd.DataFrame(self.df)

        if index:
            self.df.index = index

        return self.df


class RSTWriter():
    """ Writes a reStructured text file.

    :ivar rst: A string representing reStructured text.
    """
    def __init__(self):
        self.rst = ''

    def h1(self, txt):
        """ Adds a first level heading.

        :param txt: The text of the heading.
        """
        self.rst += txt + '\n'
        self.rst += ('=' * len(txt)) + '\n\n'

    def add(self, txt):
        """ Adds arbitrary text.

        :param txt: The text to add.
        """
        self.rst += txt

    def divider(self):
        """ Adds a divider. """
        self.rst += '\n\n----\n\n'

    def write(self, fname):
        """ Writes to a file.

        :param fname: The name of the file.
        """
        with open(fname, 'w') as f:
            f.write(self.rst)
            log_api.conf_logger(__name__).warning('Wrote to ' + fname)


# TODO read styles from file
# TODO add css params
class HTMLBuilder():
    """ Builds a HTML string.

    :ivar html: A string representing HTML.
    """
    def __init__(self):
        self.html = ''

    def h1(self, heading):
        """ Adds a first level heading.

        :param heading: The text of the heading.
        """
        style = "border-top: dashed 1px gray;"
        tag = '<h1 style="{0}">{1}</h1>'
        self.html += tag.format(style, heading)
        self.watermark()

    def h2(self, heading):
        """ Adds a second level heading.

        :param heading: The text of the heading.

        >>> from dautil import report
        >>> hb = report.HTMLBuilder()
        >>> hb.h2('Heading 2')
        >>> hb.html
        '<h2 style="color:gray;">Heading 2</h2><br/>'
        """
        style = 'color:gray;'
        self.html += '<h2 style="{0}">{1}</h2><br/>'.format(style, heading)

    def add(self, text):
        """ Adds arbitrary html.

        :param txt: The html to add.


        >>> from dautil import report
        >>> hb = report.HTMLBuilder()
        >>> hb.add('<p>42</p>')
        >>> hb.html
        '<div><p>42</p></div><br/>'
        """
        self.html += '<div>{}</div><br/>'.format(text)

    def add_df(self, df, index=False, *args, **kwargs):
        """ Converts a pandas `DataFrame` to HTML and adds it.

        :param df: A pandas `DataFrame`.
        :param index: Boolean indicating whether to display the index.
        """
        # TODO handle css properly
        classes = """
            tr:nth-child(odd)		{ background-color:#eee; }
            tr:nth-child(even)		{ background-color:#fff; }
            """

        self.html += '<div>{}</div><br/>'.format(
            df.to_html(index=index, classes=classes, *args, **kwargs))

    def __add__(self, other):
        if isinstance(other, pd.DataFrame):
            self.add_df(other)
        else:
            self.add(other)

        return self

    def watermark(self):
        """ Adds a watermark containing versions of
        detected software libraries.

        :returns: The HTML of the watermark.
        """
        versions = log_api.log(sys.modules, 'HTMLBuilder')
        style = "font-size:5px;"
        vlist = []

        for k, v in versions.items():
            vlist.append('{0}={1}'.format(k, v))

        tag = '<p style="{0}">Tested with {1}</p>'
        self.html += tag.format(style, ', '.join(vlist))

        return self.html


class Generator():
    CSS = "https://www.googledrive.com/host/0B8Dd5LtC555uNTJQXzZKdDN4aWc"

    def __init__(self, rst_file, out):
        self.dir = data.get_data_dir()
        self.out_file = os.path.join(self.dir, out)
        self.gen = LandslideGenerator(rst_file, destination_file=self.out_file)

    def add_css(self):
        css = os.path.join(self.dir, 'report.css')
        data.download(self.CSS, css)
        self.gen.add_user_css([css])

    def generate(self):
        self.add_css()
        self.gen.write()

        log_api.conf_logger(__name__).warning('Generated HTML in ' +
                                              self.out_file)



""" This module contains plotting utilities. """
from itertools import cycle
from matplotlib.colors import rgb2hex
import matplotlib.pyplot as plt
import numpy as np
from dautil import collect
from dautil import log_api
from dautil import report
from dautil import stats
from matplotlib.markers import MarkerStyle
from scipy.stats import norm


def img_show(ax, img, *args, **kwargs):
    ''' Plots an image with axes turned off.

    :param ax: A `SubplotAxes` object.
    :param img: The image to display.
    '''
    ax.imshow(img, *args, **kwargs)
    ax.axis('off')


def hist_norm_pdf(ax, arr):
    ''' Plots a histogram with corresponding normal PDF.

    :param ax: A `SubplotAxes` object.
    :param arr: Values for the histogram.
    '''
    _, bins, _ = ax.hist(arr, normed=True,
                         bins=stats.sqrt_bins(arr), label='Data')
    ax.plot(bins, norm.pdf(bins, arr.mean(),
                           arr.std()), lw=2, label='Gaussian PDF')
    ax.set_ylabel('Frequency')


def bar(ax, xlabels, vals):
    ''' Plots a bar chart.

    :param ax: A `SubplotAxes` object.
    :param xlabels: Labels on the x-axis.
    :param vals: Values for the bars.
    '''
    xpos = range(len(xlabels))
    ax.bar(xpos, vals, align='center')
    ax.set_xticks(xpos)
    ax.set_xticklabels(xlabels)


def plot_text(ax, xcoords, ycoords, point_labels, add_scatter=False,
              *args, **kwargs):
    ''' Plots text labels with given coordinates.

    :param ax: A `SubplotAxes` object.
    :param xcoords: Array-like x coordinates.
    :param ycoords: Array-like y coordinates.
    :param point_labels: Text labels to put on the chart.
    :param add_scatter: Whether to scatter plot the coordinate values.
    '''
    if add_scatter:
        ax.scatter(xcoords, ycoords)

    for x, y, txt in zip(xcoords, ycoords, point_labels):
        ax.text(x, y, txt, *args, **kwargs)


def all_markers():
    ''' Gets all the matplotlib markers except None.

    :returns: The matplotlib marker character codes except the None markers.
    '''
    return [m for m in MarkerStyle.markers.keys()
            if m is not None or m == 'None']


def map_markers(vals):
    ''' Maps matplotlib markers to values.

    :param vals: Values to map.

    :returns: A list where each value is replaced by a marker character code.
    '''
    uniqs = set(vals)
    markers = cycle(all_markers())
    mark_dict = {u: next(markers) for u in uniqs}

    return [mark_dict[v] for v in vals]


def plot_points(ax, points):
    ''' Plots points with the 'o' marker and as a line.

    :param ax: A `SubplotAxes` object to draw on.
    :param points: A list of points as the following: \
        [(x1, y1), (x2, y2), ...]
    '''
    x, y = zip(*points)
    ax.plot(x, y)
    ax.plot(x, y, 'o')


def plot_polyfit(ax, x, y, degree=1, plot_points=False):
    """ Plots a polynomial fit.

    :param ax: A matplotlib `SubplotAxes` object.
    :param x: An array of 'x' values.
    :param y: An array of 'y' values.
    :param degree: The polynomial degree.
    :param plot_points: Whether to plot points.
    """
    poly = np.polyfit(x, y, degree)

    ax.plot(x, np.polyval(poly, x), label='Fit')

    if plot_points:
        ax.plot(x, y, 'o')


def scatter_with_bar(ax, bar_label, *args, **kwargs):
    """ Creates a matplotlib scatter plot with a colorbar.

    :param ax: A matplotlib `SubplotAxes`.
    :param bar_label: The label of the colorbar.
    """
    sc = ax.scatter(*args, **kwargs)
    plt.colorbar(sc, ax=ax, label=bar_label)


def sample_cmap(name='Reds', start=0.1, end=0.9, ncolors=9):
    """ Samples a matplotlib color map
    using a linearly spaced range.

    :param name: Name of the color map.
    :param start: Start of the linear range.
    :param end: End of the linear range.
    :param ncolors: The number of colors in the range.

    :returns: A sample of the color map.

    >>> from dautil import plotting
    >>> plotting.sample_cmap()
    array([[ 0.99692426,  0.89619378,  0.84890428,  1.        ],
           [ 0.98357555,  0.41279508,  0.28835065,  1.        ],
           [ 0.59461747,  0.0461361 ,  0.07558632,  1.        ]])
    """
    cmap = plt.cm.get_cmap(name)

    return cmap(np.linspace(start, end, ncolors))


def sample_hex_cmap(name='Reds', start=0.1, end=0.9, ncolors=9):
    """ Samples a matplotlib color map
    using a linearly spaced range and
    return hex values for the colors.

    :param name: Name of the color map.
    :param start: Start of the linear range.
    :param end: End of the linear range.
    :param ncolors: The number of colors in the range.

    :returns: A list of hex values from a sample of the color map.

    >>> from dautil import plotting
    >>> plotting.sample_hex_cmap()
    ['#fee5d8', '#fdcab5', '#fcab8f', '#fc8a6a', '#fb694a',
    '#f14432', '#d92523', '#bc141a', '#980c13']
    """
    cmap = sample_cmap(name, start, end, ncolors)

    return [rgb2hex(c) for c in cmap]


def embellish(axes, legends=None):
    """ Adds grid and legends to matplotlib plots.

    :param axes: Axes as returned by the plt.subplots() function.
    :param legends: A list of indices of subplots, which need a legend.
    """
    for i, ax in enumerate(axes):
        ax.grid(True)

        if legends is None:
            ax.legend(loc='best')
        elif i in legends:
            ax.legend(loc='best')


def hide_axes(axes):
    """ Hides the x-axis and y-axis of matplotlib plots.

    :param axes: Axes as returned by the `plt.subplots()` function.
    """
    flat_axes = axes

    if len(axes) > 0:
        flat_axes = axes.ravel()

    for ax in flat_axes:
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)


class Cycler():
    """ Utility class, which cycles through values of
    plotting related lists.

    :ivar STYLES: A list of line styles.
    :ivar LW: A list of linewidths.
    :ivar colors: A list of colors.
    """
    def __init__(self, styles=["-", "--", "-.", ":"], lw=[1, 2]):
        self.STYLES = cycle(styles)
        self.LW = cycle(lw)
        self.colors = cycle(sample_hex_cmap(name='hot'))

    def style(self):
        """ Cycles through a list of line styles. """
        return next(self.STYLES)

    def lw(self):
        """ Cycles through a list of linewidth values.

        :returns: The next linewidth in the list.
        """
        return next(self.LW)

    def color(self):
        """ Cycles through a list of colors.

        :returns: The next color in the list.
        """
        return next(self.colors)


class CyclePlotter():
    """ A plotter which cycles through different linestyle
    and linewidth values.

    :ivar ax: A `SubplotAxes` instance.
    :ivar cycler: A `Cycler` instance.
    """
    def __init__(self, ax):
        self.ax = ax
        self.cycler = Cycler()

    def plot(self, x, y=None, *args, **kwargs):
        """ A facade for the matplotlib `plot()` method.

        :param x: Array of 'x' values for the plot.
        :param y: Array of 'y' values for the plot.
        """
        logger = log_api.env_logger()

        if y is None:
            y = x
            x = list(range(len(x)))

        logger.debug('len(x) %s', len(x))
        logger.debug('len(y) %s', len(y))
        self.ax.plot(x, y, self.cycler.style(),
                     lw=self.cycler.lw(), *args, **kwargs)


class Subplotter():
    """ A utility to help with subplotting.

    :ivar context: A `Context` instance.
    :ivar index: The index of the subplot.
    :ivar ax: The current `SubplotAxes` instance.
    """
    def __init__(self, nrows=1, ncols=1, context=None):
        self.context = context
        self.old = None
        self.index = -1

        if context:
            self.old = self.context.read_labels()

            if self.old:
                self.old = collect.flatten(self.old)

        # TODO turn off squeeze
        self.fig, self.axes = plt.subplots(nrows, ncols)

        if nrows > 1 and ncols > 1:
            self.axes = collect.flatten(self.axes)

        if nrows == 1 and ncols == 1:
            self.ax = self.axes
            self.index = 0
        else:
            self.ax_iter = iter(self.axes)
            self.next_ax()

    def next_ax(self):
        """ Advance to next subplot.

        :returns: The current subplot after advancing.
        """
        self.index += 1

        self.ax = next(self.ax_iter)

        return self.ax

    def get_string(self, old, key, params):
        """ Gets a string used to label x-axis,
        y-axis or title of a subplot.

        :param old: Configuration setting from a file.
        :param key: title, xlabel, legend or ylabel.
        :param params: Extra params provided for the
        Python string `format()` method. We expect the
        appropriate use of curly braces.

        :returns: A (formatted) string for the x-axis,
        y-axis, legend or title of a subplot.
        """
        astr = old.get(key, '')

        if params:
            if isinstance(params, str):
                astr = astr.format(params)
            else:
                astr = astr.format(*params)

        return astr

    def label(self, advance=False, title_params=None,
              xlabel_params=None, ylabel_params=None):
        """ Labels the subplot.

        :param advance: Boolean indicating whether to move \
            to the next subplot.
        :param title_params: Optional title parameters.
        :param xlabel_params: Optional xlabel parameters.
        :param ylabel_params: Optional ylabel parameters.
        """
        if advance:
            self.next_ax()

        # Cowardly refusing to continue
        if self.old is None:
            return

        old = self.old[self.index]

        title = self.get_string(old, 'title', title_params)

        if title:
            self.ax.set_title(title)

        xlabel = self.get_string(old, 'xlabel', xlabel_params)

        if xlabel:
            self.ax.set_xlabel(xlabel)

        ylabel = self.get_string(old, 'ylabel', ylabel_params)

        if ylabel:
            self.ax.set_ylabel(ylabel)

        legend = self.get_string(old, 'legend', None)

        if legend.startswith('loc='):
            self.ax.legend(loc=legend.replace('loc=', ''))

    # TODO Decide whether to use a context manager.
    def exit(self, mark=True):
        ''' Cleans up the `Subplotter`.

        :param mark: Boolean indicating whether to apply watermark.

        :returns: The appropriate watermark.
        '''
        plt.tight_layout()
        versions = ''

        if mark:
            versions = report.HTMLBuilder().watermark()

        return versions



""" This module contains configuration utilities. """
import json
import os


FJSON = 'dautil.json'


def file_exists(tocheck):
    """ Checks whether a file exists.

    :param tocheck: The path of the file.

    :returns: True if the file exists.
    """
    return os.path.isfile(tocheck)


def read_rc():
    """ Reads a configuration file in the JSON format.

    :returns: A dictionary representing the contents \
        of the configuration file.
    """
    old = None

    if file_exists(FJSON):
        with open(FJSON) as json_file:
            old = json.load(json_file)

    return old


def update_rc(key, updates):
    """ Updates the JSON configuration file.

    :param key: The key of the record to update.
    :param updates: Values with which to update the relevant record.
    """
    config = {key: updates}
    old = read_rc()

    if old:
        old.update(config)
        config = old

    with open(FJSON, 'w') as json_file:
        json.dump(config, json_file, indent=4, sort_keys=True)



# Temporary I think
import warnings
warnings.filterwarnings(action='ignore',
                        message='.*IPython widgets are experimental.*')
import logging

logging.basicConfig()
LOGGER = logging.getLogger(__name__)

from .nb import *
from .options import *
from .plotting import *

try:
    from .report import *
except ImportError as e:
    LOGGER.warning('Could not import report module %s', e)

from .stats import *

try:
    from .web import *
except ImportError as e:
    LOGGER.warning('Could not import web module %s', e)

try:
    from .db import *
except ImportError as e:
    LOGGER.warning('Could not import db module %s', e)

try:
    from .perf import *
except ImportError as e:
    LOGGER.warning('Could not import perf module %s', e)

try:
    from .nlp import *
except ImportError as e:
    LOGGER.warning('Could not import nlp module %s', e)



''' Utilities for web mining and HTML processing. '''
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
import urllib.parse as urlparse
import urllib.request as urlrequest


def path2url(path):
    ''' Transforms file paths to URLs starting with file:

    :param path: The file path.

    :returns: The corresponding URL.

    >>> from dautil import web
    >>> web.path2url('/home/dautil')
    'file:///home/dautil'
    '''
    return urlparse.urljoin('file:', urlrequest.pathname2url(path))


def find_hrefs(content):
    ''' Finds href links in a HTML string.

    :param content: A HTML string.

    :returns: A list of href links found by BeautifulSoup.
    '''
    soup = BeautifulSoup(content)

    return [a.get('href', '') for a in soup.findAll('a')]


def wait_browser(browser, selector, secs=10, by=By.XPATH):
    ''' Waits for a HTML element to become available.

    :param browser: An instance of a Selenium browser.
    :param selector: An expression used to select the web element.
    :param by: The selection method such as XPath or tag name.

    :returns: The web element you are waiting for.
    '''
    return WebDriverWait(browser, secs).until(
        EC.presence_of_element_located((by, selector))
    )



''' Database related utilities. '''
from dautil import log_api
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


def not_empty(session, table):
    ''' Checks whether a table is empty.

    :param session: A `SQLAlchemy` session.
    :param table: The entity class of a database table.

    :returns: `True` if the table is not empty.
    '''
    return session.query(table).count() > 0


def create_session(dbname, base, prefix='sqlite:///{}'):
    ''' Creates a database session.

    :param dbname: The name of the database.
    :param base: The declarative Base class.
    :param prefix: The string at the start of the database URL.

    :returns: The database session.
    '''
    engine = create_engine(prefix.format(dbname))
    DBSession = sessionmaker(bind=engine)
    base.metadata.create_all(engine)

    return DBSession()


def count_where(session, column, search_val):
    ''' Gets the count of a query as in the following SQL \
        `SELECT count(column) from atable where atable = 'search_val'`\

        .. note:: **DO NOT USE** when high-performance is required.

    :param session: A `SQLAlchemy` session.
    :param column: A table column as attribute in a Python class.
    :param search_val: Value to search for.

    :returns: The count of the query.
    '''
    return session.query(column).filter(column == search_val).count()


def entity_from_column(column):
    ''' Utility function which returns the entity of a column.

    :param column: A table column as attribute in a Python class.

    :returns: The entity.
    '''
    return column.parent.entity


def where_first(session, column, search_val):
    ''' Gets the first row as in the following SQL \
        `SELECT * from atable where atable = 'search_val'`

    :param session: A `SQLAlchemy` session.
    :param column: A table column as attribute in a Python class.
    :param search_val: Value to search for.

    :returns: The first row of the query.
    '''
    logger = log_api.env_logger()
    result = session.query(entity_from_column(column)).\
        filter(column == search_val).first()
    logger.debug('search_val=%s result=%s', search_val, result)

    return result


def map_to_id(session, column):
    ''' Assuming that a table has a primary key column 'id'. \
        This function creates dictionary with column value as key \
        and id as value.

    :param session: A `SQLAlchemy` session.
    :param column: A table column as attribute in a Python class.

    :returns: A dictionary with column values mapped to id.
    '''
    id_map = {}
    logger = log_api.env_logger()
    table = column.parent.entity
    query = session.query(table.id, column)
    logger.debug('Query %s', query)

    for id, val in query.all():
        id_map[val] = id

    logger.debug('Id map %s', id_map)

    return id_map



""" This module contains utilities related to collections."""

from dautil import log_api
from itertools import chain
from itertools import groupby
import collections


def chunk(alist, size):
    """Yields equal sized chunks from a list.

    :param alist: List-like sequence.
    :param size: Size of the chunk.

    >>> from dautil import collect
    >>> collect.chunk(list(range(7)), 3)
    <generator object chunk at 0x1006a7480>
    >>> next(collect.chunk(list(range(7)), 3))
    [0, 1, 2]
    """
    for i in range(0, len(alist), int(size)):
        yield alist[i:i + size]


def longest_streak(arr, val):
    ''' Counts the number of elements in the longest streak \
        for a given value.

    :param arr: A 1-d list.
    :param val: The value to search for.

    :returns: The count of the longest sequence.

    >>> from dautil import collect
    >>> collect.longest_streak([0, 1, 1, 0, 1, 1, 1], 1)
    3
    '''
    max_count = 0

    for k, g in groupby(arr, lambda x: x == val):
        if not k:
            continue

        count = sum(1 for i in g)

        if count > max_count:
            max_count = count

    return max_count


def sort_dict_by_keys(adict):
    """ Sorts a dictionary by keys.

    :param adict: A dictionary.

    :returns: An `OrderedDict` sorted by keys.

    >>> from dautil import collect
    >>> food = {'spam': 42, 'eggs': 28}
    >>> collect.sort_dict_by_keys(food)
    OrderedDict([('eggs', 28), ('spam', 42)])
    """
    return collections.OrderedDict(sorted(adict.items(), key=lambda t: t[0]))


def dict_from_keys(adict, keys):
    """ Selects a subset of a dictionary with a list of keys.

    :param adict: A dictionary.
    :param keys: A list of keys.

    :returns: A subset of the input dictionary.

    >>> from dautil import collect
    >>> adict = {'a.latex': 1, 'b.latex': 2, 'c': 3}
    >>> collect.dict_from_keys(adict, ['b.latex', 'a.latex'])
    {'a.latex': 1, 'b.latex': 2}
    """
    return {k: adict[k] for k in keys}


def filter_list(func, alist):
    """ Filters a list using a function.

    :param func: A function used for filtering.
    :param alist: The list to filter.

    :returns: The filtered list.

    >>> from dautil import collect
    >>> alist = ['a', 'a.color', 'color.b']
    >>> collect.filter_list(lambda x: x.endswith('color'), alist)
    ['a.color']
    """
    return [a for a in alist if func(a)]


def filter_dict_keys(func, adict):
    """ Filters the keys of a dictionary.

    :param func: A function used to filter the keys.
    :param adict: A dictionary.

    :returns: A list of keys selected by the filter.

    >>> from dautil import collect
    >>> adict = {'a.latex': 1, 'b.latex': 2, 'c': 3}
    >>> collect.filter_dict_keys(lambda x: x.endswith('.latex'), adict)
    ['a.latex', 'b.latex']
    """
    logger = log_api.env_logger()
    logger.debug('adict {}'.format(adict))

    return [k for k in adict.keys() if func(k)]


def dict_updates(old, new):
    """ This function reports updates to a dict.

    :param old: A dictionary to compare against.
    :param new: A dictionary with potential changes.

    :returns: A dictionary with the updates if any.
    """
    updates = {}

    for k in set(new.keys()).intersection(set(old.keys())):
        if old[k] != new[k]:
            updates.update({k: new[k]})

    return updates


def is_rectangular(alist):
    """ Checks whether a list is rectangular.

    :param alist: A list or similar data structure.

    :returns: True if the argument is rectangular.

    >>> from dautil import collect
    >>> collect.is_rectangular([[2, 1], [3]])
    False
    >>> collect.is_rectangular([[2, 1], [3, 4]])
    True
    """
    lengths = {len(i) for i in alist}

    return len(lengths) == 1


def isiterable(tocheck):
    """ Checking for an iterable argument using a
    somewhat modified definition ie strings and dicts
    are considered not iterable.

    :param tocheck: The data structure to check.

    :returns: True if iterable
    """
    return not isinstance(tocheck, str) and\
        not isinstance(tocheck, dict) and\
        isinstance(tocheck, collections.Iterable)


def flatten(iterable):
    """ Flattens an iterable, where strings and dicts
    are not considered iterable.

    :param iterable: The iterable to flatten.

    :returns: The iterable flattened as a flat list.

    >>> from dautil import collect
    >>> collect.flatten([[1, 2]])
    [1, 2]
    """
    logger = log_api.env_logger()
    logger.debug('Iterable {}'.format(iterable))
    assert isiterable(iterable), 'Not iterable {}'.format(iterable)
    flat = iterable

    if isiterable(iterable[0]):
        flat = [i for i in chain.from_iterable(iterable)]

    return flat


def grid_list(arr):
    ''' Creates a 2D cartesian product grid from an array:

    >>> from dautil import collect
    >>>  arr = list(range(2))
    >>> arr
    [0, 1]
    >>> collect.grid_list(arr)
    [[(0, 0), (0, 1)], [(1, 0), (1, 1)]]


    :param arr: A 1-d array-like list.

    :returns: A 2-d list containing tuples as elements.
    '''
    dim = len(arr)
    grid = GridList(dim, dim, 0).grid

    for i in range(dim):
        for j in range(dim):
            grid[i][j] = (arr[i], arr[j])

    return grid


class GridList():
    """ A two-dimensional rectangular list.

    :ivar nrows: The number of rows in the grid.
    :ivar ncols: The number of columns in the grid.
    :ivar val: The initial value of cells in the grid.
    :ivar logger: The logger of the class.
    """
    def __init__(self, nrows, ncols, val):
        self.nrows = nrows
        self.ncols = ncols
        self.grid = [ncols * [val] for i in range(nrows)]
        self.logger = log_api.conf_logger('collect.GridList')

    def dim_equal(self, other):
        """ Checks whether the dimensions of the current GridList
        and another rectangular list are equal.

        :param other: Another rectangular list.

        :returns: True if both data structures have the same dimensions.
        """
        return len(other) == self.nrows and len(other[0]) == self.ncols

    def copy(self, other):
        """ Copies the contents of another list
        into the current GridList.

        :param other: Another list.
        """
        flat_grid = flatten(self.grid)
        flat_other = chain.from_iterable(other)

        for i, value in enumerate(flat_other):
            flat_grid[i] = value

        item = iter(flat_grid)

        for i in range(self.nrows):
            for j in range(self.ncols):
                self.grid[i][j] = next(item)

    def fill(self, other):
        """ Fills the current GridList with the contents
        of another GridList as much as possible, unless
        there is a chance of overflow.

        :param: other: Another list.
        """
        # If edited may become jagged
        if is_rectangular(other):
            if self.dim_equal(other):
                self.grid = other

                return

            ocells = len(other) * len(other[0])
            ncells = self.nrows * self.ncols

            if ocells > ncells:
                # might lose info
                self.logger.warning('Filling {0} with {1}'.format(ncells,
                                                                  ocells))
            else:
                self.copy(other)

    def update(self, row, col, change):
        """ Updates a specific grid cell.

        :param row: The row number of the cell.
        :param col: The column number of the cell.
        :param change: The content with which to update\
            the cell. If the cell is a dict and the update value\
            too then the original dict will be updated.
        """
        cell = self.grid[row][col]

        if isinstance(cell, dict):
            if len(cell.keys()) == 0:
                cell = change
            else:
                (cell).update(change)

        self.grid[row][col] = cell


class IdDict():
    ''' Dictionary for items identified by integers.

    :ivar curr: The current position.
    :ivar items: The dictionary.
    '''
    def __init__(self, items={}):
        self.curr = 0
        self.items = items

    def add_or_get(self, item):
        ''' Adds or gets the id of an item.

        :param item: The item.

        :returns: The id.
        '''
        if self.items.get(item, None) is None:
            self.curr += 1
            self.items[item] = self.curr

        return self.items[item]

    def get_id(self, item):
        ''' Gets the id of an item.

        :param item: The item.

        :returns: The id.
        '''
        return self.items[item]



""" Statistical functions and utilities. """
import numpy as np
from scipy.stats import describe
import pandas as pd
import math
from collections import namedtuple
from math import sqrt


def rss(actual, forecast):
    ''' Computes the residuals squares sum (RSS).

    .. math:: RSS = \sum_{i=1}^n (y_i - f(x_i))^2

    :param actual: The target values.
    :param forecast: Predicted values.

    :returns: The RSS.
    '''
    return ((actual - forecast) ** 2).sum()


def mse(actual, forecast):
    ''' Computes the mean squared error.

    .. math:: \operatorname{MSE}=\\frac{1}{n}\\sum_{i=1}^n(\\hat{Y_i} - Y_i)^2

    :param actual: The target values.
    :param forecast: Predicted values.

    :returns: The MSE.

    '''
    return ((actual - forecast) ** 2).mean()


def mpe(actual, forecast):
    ''' Computes the mean percentage error. \
        Non-finite values due to zero division are ignored.

    .. math:: \\text{MPE} = \\frac{100\\%}{n}\\sum_{t=1}^n \\frac{a_t-f_t}{a_t}

    :param actual: The target values.
    :param forecast: Predicted values.

    :returns: The MPE.
    '''
    ratio = (actual - forecast)/actual
    ratio = ratio[np.isfinite(ratio)]

    return 100 * ratio.mean()


def mape(actual, forecast):
    ''' Computes the mean absolute percentage error. \
        Non-finite values due to zero division are ignored.

    .. math::  \mbox{MAPE} = \\frac{1}{n}\\sum_{t=1}^n  \
        \\left|\\frac{A_t-F_t}{A_t}\\right|

    :param actual: The target values.
    :param forecast: Predicted values.

    :returns: The MAPE.
    '''
    ratio = (actual - forecast)/actual
    ratio = ratio[np.isfinite(ratio)]

    return np.abs(ratio).mean()


def wssse(point, center):
    ''' Computes the **Within Set Sum of Squared Error(WSSSE)**.

    :param point: A point for which to calculate the error.
    :param center: The center of a cluster.

    :returns: The WSSSE.
    '''
    return sqrt(sum([x**2 for x in (point - center)]))


def trimean(arr):
    """ Calculates the trimean.

    .. math:: TM=\\frac{Q_1 + 2Q_2 + Q_3}{4}

    :param arr: An array containing numbers.

    :returns: The trimean for the array.

    >>> import numpy as np
    >>> from dautil import stats
    >>> stats.trimean(np.arange(9))
    4.0

    """
    q1 = np.percentile(arr, 25)
    q2 = np.percentile(arr, 50)
    q3 = np.percentile(arr, 75)

    return (q1 + 2 * q2 + q3)/4


def outliers(arr, method='IQR', factor=1.5, percentiles=(5, 95)):
    """ Gets the limits given an array for values to be considered
    outliers.

    :param arr: An array containing numbers.
    :param method: IQR (default) or percentiles.
    :param factor: Factor for the IQR method.
    :param percentiles: A tuple of percentiles.

    :returns: A namedtuple with upper and lower limits.

    >>> import numpy as np
    >>> from dautil import stats
    >>> stats.outliers(a)
    Outlier(min=-48.5, max=149.5)
    """
    Outlier = namedtuple('Outlier', ['min', 'max'])
    outlier = None

    if method == 'IQR':
        q1 = np.percentile(arr, 25)
        q3 = np.percentile(arr, 75)
        iqr = q3 - q1
        outlier = Outlier(q1 - factor * iqr, q3 + factor * iqr)
    elif method == 'percentiles':
        minmax = [np.percentile(arr, p) for p in percentiles]
        outlier = Outlier(minmax[0], minmax[1])
    else:
        raise AssertionError('Unknown method expected IQR or percentiles')

    return outlier


def clip_outliers(arr):
    """ Clips values using limits for values
    to be considered outliers.

    :param arr: An array containing numbers.

    :returns: A clipped values array.

    >>> import numpy as np
    >>> from dautil import stats
    >>> arr = list(range(5))
    >>> outliers = [-100, 100]
    >>> arr.extend(outliers)
    >>> arr
    [0, 1, 2, 3, 4, -100, 100]
    >>> stats.clip_outliers(arr)
    array([ 0.,  1.,  2.,  3.,  4., -4.,  8.])
    """
    amin, amax = outliers(arr)

    return np.clip(arr, amin, amax)


def zscores(x):
    return (x - x.mean())/x.std()


def sqrt_bins(arr):
    """ Uses a rule of thumb to calculate
    the appropriate number of bins for an array.

    :param arr: An array containing numbers.

    :returns: An integer to serve as the number of bins.
    """
    return int(math.sqrt(len(arr)))


def ci(arr, alpha=0.95):
    """ Computes the confidence interval.

    :param arr: An array containing numbers.
    :param alpha: A value in the range 0 - 1\
    that serves as percentiles.

    :returns: The confidence interval.
    """
    interval = 1 - alpha

    minmax = [interval/2, alpha + interval/2]

    return np.percentile(arr, 100 * np.array(minmax))


def jackknife(arr, func, alpha=0.95):
    """ Jackknifes an array with a supplied function.

    :param arr: An array containing numbers.
    :param func: The function to apply.
    :param alpha: A number in the range 0 - 1 used\
    to calculate the confidence interval.

    :returns: Three numbers - the function value for arr,\
    the lower limit of the confidence interval and the\
    upper limit of the confidence interval.
    """
    n = len(arr)
    idx = np.arange(n)

    low, high = ci([func(arr[idx != i]) for i in range(n)], alpha)

    return func(arr), low, high


class Distribution():
    """ Wraps `scipy.stats` distribution classes.
    Most of the methods are only appropriate for continuous distributions.

    :ivar nbins: The number of bins for a histogram plot.
    :ivar train: The train data.
    :ivar test: The test data.
    """
    def __init__(self, data, dist, nbins=20, cutoff=0.75, range=None):
        self.nbins = nbins
        self.train, self.test = self.split(data, cutoff)
        self.hist_values, edges = np.histogram(self.test, bins=self.nbins,
                                               range=range, density=True)
        self.x = 0.5 * (edges[1:] + edges[:-1])
        self.dist = dist
        self.residuals = None
        self.pdf_values = None

    def describe_residuals(self, *args, **kwds):
        """ Describes the residuals of a fit to a distribution.
        Only appropriate for continuous distributions.

        :returns: Statistics for the residuals as a dict.
        """
        _, _, mean, var, skew, kurtosis = describe(self.error(*args, **kwds))
        result = {}
        result['Mean'] = mean
        result['Var'] = var
        result['Skew'] = skew
        result['Kurtosis'] = kurtosis
        result = pd.DataFrame([result])
        result.index = ['Residuals Statistics']

        return result

    def error(self, *args, **kwds):
        """ Computes the residuals of a fit to a distribution.
        Only appropriate for continuous distributions.

        :returns: The residuals of the fit.
        """
        if self.residuals is None:
            self.residuals = self.hist_values - self.pdf(*args, **kwds)

        return self.residuals

    def fit(self, *args):
        """ Fits data to a distribution.
        Only appropriate for continuous distributions.

        :returns: The result of the fit.
        """
        return self.dist.fit(self.train, *args)

    def mean(self):
        return self.train.mean()

    def mean_ad(self, *args, **kwds):
        """ Computes the mean absolute deviation.

        :returns: The mean absolute deviation.
        """
        abserror = np.abs(self.error(*args, **kwds))

        return abserror.mean()

    def pdf(self, *args, **kwds):
        """ Computes the probability distribution function.
        Only appropriate for continuous distributions.

        :returns: The probability distribution function.
        """
        if self.pdf_values is None:
            self.pdf_values = self.dist.pdf(self.x, *args, **kwds)

        return self.pdf_values

    def plot(self, ax):
        """ Plots a histogram of the data and a fit.
        Only appropriate for continuous distributions.
        """
        ax.hist(self.train, bins=self.nbins, normed=True, label='Data')
        ax.plot(self.x, self.pdf_values, label='PDF')
        ax.set_ylabel('Probability')
        ax.legend(loc='best')

    def rmse(self, *args, **kwds):
        """ Computes the root mean square error of the distribution fit.
        Only appropriate for continuous distributions.

        :returns: The RMSE of the fit.
        """
        se = self.error(*args, **kwds) ** 2

        return np.sqrt(se.mean())

    def rvs(self, *args, **kwds):
        """ Generates random values.

        :returns: The generated values.
        """
        return self.dist.rvs(*args, **kwds)

    def split(self, data, cutoff):
        """ Splits data into test and train data.

        :param data: An array containing numbers.
        :param cutoff: A value in the range 0 - 1 used\
        to split the data.

        :returns: Two arrays - the train data and the test data.
        """
        n = len(data)
        n = int(cutoff * n)

        return data[:n], data[n:]

    def var(self):
        return self.train.var()


# TODO Decide on interpolation
class Box():
    ''' Implements the concept of a box we know from box plots. '''
    def __init__(self, arr):
        self.arr = arr
        self.q1 = np.percentile(arr, 25, interpolation='nearest')
        self.q3 = np.percentile(arr, 75, interpolation='nearest')

    def calc_iqr(self):
        """ Calculates the Inter Quartile Range.

        :returns: The interquartile range of the data.

        >>> import numpy as np
        >>> from dautil import stats
        >>> arr = np.array([7, 15, 36, 39, 40, 41])
        >>> box = stats.Box(arr)
        >>> box.calc_iqr()
        25
        """
        return self.q3 - self.q1

    def iqr_from_box(self):
        """ Calculates the number (0 or more) of
        inter quartile ranges (IQR) from the box formed
        by the first and third quartile.

        :returns: A float representing the number of IQRs
        """
        iqr = self.calc_iqr()
        from_q1 = np.abs(self.arr - self.q1)/iqr
        from_q3 = np.abs(self.arr - self.q3)/iqr

        return np.minimum(from_q1, from_q3)



from setuptools import setup

setup(name='dautil',
      version='0.0.1a32',
      description='Data analysis utilities',
      url='https://www.packtpub.com/books/info/authors/ivan-idris',
      author='Ivan Idris',
      author_email='ivan.idris@gmail.com',
      license='MIT',
      packages=['dautil'],
      include_package_data=True,
      test_suite='nose.collector',
      tests_require=['nose', 'nose-parameterized'],
      install_requires=[
          'appdirs',
          'landslide'
      ],
      zip_safe=False)



""" Configures dynamic options of several
data analysis related libraries. """
import pandas as pd
import matplotlib as mpl
import numpy as np


APP_DIR = 'dautil'


def mimic_seaborn():
    """ Mimics the Seaborn style with small deviations. """
    sns_style = {'axes.axisbelow': True,
                 'axes.edgecolor': 'white',
                 'axes.facecolor': '#EAEAF2',
                 'axes.grid': True,
                 'axes.labelcolor': '.15',
                 'axes.linewidth': 0,
                 'font.family': 'Arial',
                 # deviating
                 'grid.color': 'yellow',
                 'grid.linestyle': '-',
                 # deviating
                 'grid.linewidth': 2,
                 'image.cmap': 'Greys',
                 'legend.frameon': False,
                 'legend.numpoints': 1,
                 'legend.scatterpoints': 1,
                 'lines.solid_capstyle': 'round',
                 'pdf.fonttype': 42,
                 'text.color': '.15',
                 'xtick.color': '.15',
                 'xtick.direction': 'out',
                 'xtick.major.size': 0,
                 'xtick.minor.size': 0,
                 'ytick.color': '.15',
                 'ytick.direction': 'out',
                 'ytick.major.size': 0,
                 'ytick.minor.size': 0}
    mpl.rcParams.update(sns_style)


def set_pd_options():
    """ Sets pandas options. """
    pd.set_option('precision', 4)
    pd.set_option('max_rows', 5)


def reset_pd_options():
    """ Resets pandas options. """
    pd.reset_option('precision')
    pd.reset_option('max_rows')


def set_mpl_options():
    """ Sets matplotlib options. """
    mpl.rcParams['legend.fancybox'] = True
    mpl.rcParams['legend.shadow'] = True
    mpl.rcParams['legend.framealpha'] = 0.7


def set_np_options():
    """ Sets NumPy options. """
    np.set_printoptions(precision=4, threshold=5,
                        linewidth=65)


def reset_np_options():
    """ Resets NumPy options. """
    np.set_printoptions(precision=8, threshold=1000,
                        linewidth=75)



""" Utilities for time series and dates. """
import pandas as pd
import calendar
import numpy as np
from scipy import signal


def ar1(arr):
    ''' Fits AR(1) model.

    :param arr: array-like, shape(M,)

    :returns A dictionary with slope and intercept of the model.
    '''
    slope, intercept = np.polyfit(arr[:-1], arr[1:], 1)

    return {'slope': slope, 'intercept': intercept}


def power(arr):
    ''' Computes the power of a signal.

    :param arr: Array-like list of values.

    :returns: The power of a signal.
    '''
    return np.abs(arr) ** 2


def instant_phase(arr):
    ''' Computes the instaneous phase of a signal.

    :param arr: Array-like list of values.

    :returns: The instaneous phase.
    '''
    return np.angle(signal.hilbert(arr))


def double_exp_smoothing(arr, alpha, beta):
    ''' Applies double exponential smoothing.

    .. math::
        \\begin{align}
        s_{t} = \\alpha x_{t} + (1-\\alpha)(s_{t-1} + b_{t-1})\\\\
        b_{t} = \\beta (s_t - s_{t-1}) + (1-\\beta)b_{t-1}\\\\
        \\end{align}

    :param arr: Array-like list of values.
    :param alpha: The smoothing factor parameter.
    :param beta: The trend factor parameter.

    :returns: The smoothed series values.
    '''
    st = arr.copy()
    bt = arr.copy()
    bt[0] = (arr[-1] - arr[0])/len(arr)
    bt[1] = arr[1] - arr[0]

    for i, val in enumerate(arr[1:], start=1):
        st[i] = alpha * val + (1 - alpha) * (st[i - 1] + bt[i - 1])
        bt[i] = beta * (st[i] - st[i - 1]) + (1 - beta) * bt[i - 1]

    return st


def exp_smoothing(arr, alpha):
    ''' Applies exponential smoothing.

    .. math:: s_t = \\alpha \\cdot x_{t} + (1-\\alpha) \\cdot s_{t-1}

    :param arr: Array-like list of values.
    :param alpha: The smoothing factor parameter.

    :returns: The smoothed series values.
    '''
    st = arr.copy()

    for i, val in enumerate(arr[1:], start=1):
        st[i] = alpha * val + (1 - alpha) * st[i - 1]

    return st


def rolling_deviations(arr, window):
    ''' Computes the rolling deviations of a series, by subtracting \
        the rolling mean and dividing by the rolling standard deviation.

    :param arr: Array-like list of values.
    :param window: Size of the window.

    :returns: The rolling deviations.
    '''
    return (arr - pd.rolling_mean(arr, window))/pd.rolling_std(arr, window)


def fano_factor(arr, window):
    ''' Calculates the Fano factor a windowed variance-to-mean ratio.

    .. math:: F=\\frac{\\sigma_W^2}{\\mu_W}

    :param arr: Array-like list of values.
    :param window: Size of the window.

    :returns: The Fano factor.

    *See Also*

    https://en.wikipedia.org/wiki/Fano_factor
    '''
    return pd.rolling_var(arr, window)/pd.rolling_mean(arr, window)


def sine_like(arr):
    ''' Creates a sine wave of roughly the same size as the input.

    :param arr: Array-like list of values.

    :returns: A sine wave.
    '''
    t = np.linspace(-2 * np.pi, 2 * np.pi, len(arr))
    mid = np.ptp(arr)/2

    return mid + mid * np.sin(t)


def groupby_yday(df):
    """ Groups a pandas DataFrame by the day of year.

    :param df: A pandas `DataFrame`.

    :returns: The grouped `DataFrame`.
    """
    return df.groupby(lambda d: d.timetuple().tm_yday)


def groupby_month(df):
    """ Groups a pandas `DataFrame` by month.

    :param df: A pandas `DataFrame`.

    :returns: The grouped `DataFrame`.
    """
    return df.groupby(df.index.month)


def groupby_year(df):
    """ Groups a pandas `DataFrame` by year.

    :param df: A pandas `DataFrame`.

    :returns: The grouped `DataFrame`.
    """
    return df.groupby(df.index.year)


def groupby_year_month(df):
    """ Groups a pandas `DataFrame` by year and month.

    :param df: A pandas `DataFrame`.

    :returns: The grouped `DataFrame`.
    """
    return pd.groupby(df, by=[df.index.year, df.index.month])


def short_month(i, zero_based=False):
    """ Looks up the short name of a month with an index.

    :param i: Index of the month to lookup.
    :param zero_based: Indicates whether the index starts from 0.

    :returns: The short name of the month for example Jan.

    >>> from dautil import ts
    >>> ts.short_month(1)
    'Jan'
    """
    j = i

    if zero_based:
        if i < 0 or i > 11:
            raise AssertionError("Out of range " + i)

        j = i + 1

    return calendar.month_abbr[j]


def short_months():
    """ Gets the short names of the months.

    :returns: A list containing the short month names.
    """
    return [short_month(i) for i in range(13)]


def month_index(month, zero_based=False):
    """ Looks up the index of a month from a short name.

    :param month: The short name of a month to lookup for example Jan.
    :param zero_based: Indicates whether the index starts from 0.

    :returns: The index of the month.

    >>> from dautil import ts
    >>> ts.month_index('Jan')
    1
    """
    for i, m in enumerate(short_months()):
        if m == month:
            index = i

            if not zero_based:
                index += 1

            return i
